package com.atlassian.elasticsearch.buckler;

import com.atlassian.elasticsearch.buckler.config.AuthConfig;
import com.atlassian.elasticsearch.buckler.config.BucklerConfig;
import com.atlassian.elasticsearch.buckler.security.AuthRateLimiter;
import com.atlassian.elasticsearch.buckler.security.RequestIdentifier;
import org.elasticsearch.common.util.concurrent.ThreadContext;
import org.elasticsearch.tasks.Task;
import org.elasticsearch.transport.TransportChannel;
import org.elasticsearch.transport.TransportRequest;
import org.elasticsearch.transport.TransportRequestHandler;

class AuthTransportRequestHandler<T extends TransportRequest> implements TransportRequestHandler<T> {

    private final AuthRateLimiter authRateLimiter;
    private final ThreadContext threadContext;
    private final BucklerConfig config;
    private final TransportRequestHandler<T> delegate;

    AuthTransportRequestHandler(TransportRequestHandler<T> delegate, BucklerConfig config,
                                AuthRateLimiter authRateLimiter, ThreadContext threadContext) {
        this.delegate = delegate;
        this.config = config;
        this.authRateLimiter = authRateLimiter;
        this.threadContext = threadContext;
    }

    @Override
    public void messageReceived(T request, TransportChannel channel, Task task) throws Exception {
        if (allow(request)) {
            delegate.messageReceived(request, channel, task);
        } else {
            throw new IllegalStateException("Authentication error");
        }
    }

    @Override
    public void messageReceived(T request, TransportChannel channel) throws Exception {
        if (allow(request)) {
            delegate.messageReceived(request, channel);
        } else {
            throw new IllegalStateException("Authentication error");
        }
    }

    private boolean allow(T request) {
        AuthConfig authConfig = config.getAuthConfig();
        if (!authConfig.isEnabledForTcp()) {
            return true;
        }

        RequestIdentifier requestIdentifier = RequestIdentifier.from(request);

        /**
         * In the event that we don't already trust this IP and we're in a client or global lockout we're going to
         * instantly fail here. This means that we do not check the password. This will protect from timing attacks
         * as there is no difference between a failed request with the correct password and a failed request with
         * an incorrect password.
         */
        if (!authRateLimiter.isRequestAllowed(requestIdentifier)) {
            return false;
        }

        String authorization = threadContext.getHeader("Authorization");

        if (authorization == null) {
            return false;
        }

        if (authConfig.isAuthorized(authorization)) {
            return true;
        } else {
            authRateLimiter.registerFailure(requestIdentifier);
            return false;
        }
    }
}
