package com.atlassian.elasticsearch.buckler;

import com.atlassian.elasticsearch.buckler.config.BucklerConfig;
import com.atlassian.elasticsearch.buckler.config.ConfigRestHandler;
import com.atlassian.elasticsearch.buckler.security.AuthRateLimiter;
import com.atlassian.elasticsearch.buckler.security.TimeProvider;
import org.elasticsearch.cluster.metadata.IndexNameExpressionResolver;
import org.elasticsearch.cluster.node.DiscoveryNodes;
import org.elasticsearch.common.inject.Inject;
import org.elasticsearch.common.io.stream.NamedWriteableRegistry;
import org.elasticsearch.common.network.NetworkService;
import org.elasticsearch.common.settings.*;
import org.elasticsearch.common.util.BigArrays;
import org.elasticsearch.common.util.PageCacheRecycler;
import org.elasticsearch.common.util.concurrent.ThreadContext;
import org.elasticsearch.common.xcontent.NamedXContentRegistry;
import org.elasticsearch.http.HttpServerTransport;
import org.elasticsearch.indices.breaker.CircuitBreakerService;
import org.elasticsearch.plugins.ActionPlugin;
import org.elasticsearch.plugins.NetworkPlugin;
import org.elasticsearch.plugins.Plugin;
import org.elasticsearch.rest.RestController;
import org.elasticsearch.rest.RestHandler;
import org.elasticsearch.threadpool.ThreadPool;
import org.elasticsearch.transport.Transport;
import org.elasticsearch.transport.TransportInterceptor;

import java.io.IOException;
import java.nio.file.Path;
import java.util.*;
import java.util.function.Supplier;
import java.util.function.UnaryOperator;

/**
 * Buckler security plugin for basic authentication and TLS.
 */
public class BucklerPlugin extends Plugin implements NetworkPlugin, ActionPlugin {

    private static final Setting<Boolean> AUTH_BASIC_HTTP_ENABLED_SETTING =
            Setting.boolSetting("auth.basic.http.enabled", false, Setting.Property.NodeScope);
    private static final Setting<Boolean> AUTH_BASIC_TCP_ENABLED_SETTING =
            Setting.boolSetting("auth.basic.tcp.enabled", false, Setting.Property.NodeScope);
    private static final Setting<String> AUTH_BASIC_USERNAME_SETTING =
            Setting.simpleString("auth.basic.username", Setting.Property.NodeScope);
    private static final Setting<String> AUTH_BASIC_PASSWORD_SETTING =
            Setting.simpleString("auth.basic.password", Setting.Property.NodeScope);
    private static final Setting<Boolean> TLS_HTTP_ENABLED_SETTING =
            Setting.boolSetting("tls.http.enabled", false, Setting.Property.NodeScope);
    private static final Setting<Boolean> TLS_TCP_ENABLED_SETTING =
            Setting.boolSetting("tls.tcp.enabled", false, Setting.Property.NodeScope);
    private static final Setting<String> TLS_KEYSTORE_PATH_SETTING =
            Setting.simpleString("tls.keystore.path", Setting.Property.NodeScope);
    private static final Setting<String> TLS_KEYSTORE_PASSWORD_SETTING =
            Setting.simpleString("tls.keystore.password", Setting.Property.NodeScope);

    private final AuthRateLimiter limiter;
    private final BucklerConfig bucklerConfig;

    @Inject
    public BucklerPlugin(Settings settings, Path configPath) throws IOException {
        this.bucklerConfig = new BucklerConfig(settings, configPath);
        this.limiter = new AuthRateLimiter(bucklerConfig, new TimeProvider());
    }

    @Override
    public Map<String, Supplier<HttpServerTransport>> getHttpTransports(Settings settings, ThreadPool threadPool,
                                                                        BigArrays bigArrays, CircuitBreakerService circuitBreakerService,
                                                                        NamedWriteableRegistry namedWriteableRegistry,
                                                                        NamedXContentRegistry xContentRegistry,
                                                                        NetworkService networkService, HttpServerTransport.Dispatcher dispatcher) {
        return Collections.singletonMap("buckler", () -> new SecureHttpServerTransport(settings, networkService, bigArrays,
                bucklerConfig, threadPool, xContentRegistry, dispatcher));
    }

    @Override
    public UnaryOperator<RestHandler> getRestHandlerWrapper(ThreadContext threadContext) {
        return restHandler -> new RestAuthenticationHandler(restHandler, bucklerConfig, limiter);
    }

    @Override
    public List<RestHandler> getRestHandlers(Settings settings, RestController restController,
                                             ClusterSettings clusterSettings, IndexScopedSettings indexScopedSettings,
                                             SettingsFilter settingsFilter, IndexNameExpressionResolver indexNameExpressionResolver,
                                             Supplier<DiscoveryNodes> nodesInCluster) {
        List<RestHandler> arrayList = new ArrayList<>();
        arrayList.add(new ConfigRestHandler(settings, restController, bucklerConfig));
        return arrayList;
    }

    @Override
    public List<Setting<?>> getSettings() {
        ArrayList<Setting<?>> settings = new ArrayList<>();
        settings.add(AUTH_BASIC_HTTP_ENABLED_SETTING);
        settings.add(AUTH_BASIC_PASSWORD_SETTING);
        settings.add(AUTH_BASIC_TCP_ENABLED_SETTING);
        settings.add(AUTH_BASIC_USERNAME_SETTING);
        settings.add(TLS_HTTP_ENABLED_SETTING);
        settings.add(TLS_KEYSTORE_PASSWORD_SETTING);
        settings.add(TLS_KEYSTORE_PATH_SETTING);
        settings.add(TLS_TCP_ENABLED_SETTING);
        return settings;
    }

    @Override
    public List<TransportInterceptor> getTransportInterceptors(NamedWriteableRegistry namedWriteableRegistry, ThreadContext threadContext) {
        return Collections.singletonList(new AuthTransportInterceptor(limiter, bucklerConfig, threadContext));
    }

    @Override
    public Map<String, Supplier<Transport>> getTransports(Settings settings, ThreadPool threadPool,
                                                          PageCacheRecycler pageCacheRecycler,
                                                          CircuitBreakerService circuitBreakerService,
                                                          NamedWriteableRegistry namedWriteableRegistry,
                                                          NetworkService networkService) {
        return Collections.singletonMap("buckler", () -> new SecureTransport(settings, threadPool, networkService,
                pageCacheRecycler, bucklerConfig, namedWriteableRegistry, circuitBreakerService));
    }

}
