package com.atlassian.elasticsearch.buckler.security;

import com.atlassian.elasticsearch.buckler.config.BruteForceProtectionSettings;
import com.atlassian.elasticsearch.buckler.config.BucklerConfig;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.elasticsearch.common.inject.Inject;
import org.elasticsearch.common.inject.Singleton;

import java.util.Calendar;
import java.util.Date;
import java.util.Iterator;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.ConcurrentSkipListSet;

@Singleton
public class AuthRateLimiter {
    private static final Logger log = LogManager.getLogger(AuthRateLimiter.class.getName());

    private final ConcurrentMap<String, AuthRecord> authenticationRecords;
    private final BruteForceProtectionSettings bruteForceConfig;
    private final TimeProvider timeProvider;
    private final Set<String> trustedHosts;

    private volatile boolean globalLockout;

    @Inject
    public AuthRateLimiter(BucklerConfig config, TimeProvider timeProvider) {
        this.trustedHosts = new ConcurrentSkipListSet<>();
        authenticationRecords = new ConcurrentHashMap<>();
        globalLockout = false;
        this.timeProvider = timeProvider;
        bruteForceConfig = config.getBruteForceProtectionSettings();
    }

    public void addTrustedRequest(RequestIdentifier request) {
        if (!request.isAnonymous()) {
            log.debug("{} has been added as a trusted host", request.getIdentifier());
            trustedHosts.add(request.getIdentifier());
        }
    }

    /**
     * Checks if the client is allowed to perform a request to ES. If false, no authorization should be checked
     * and clients request should be rejected as unauthorised.
     *
     * @param request the client request request
     * @return true if the client is either trusted, or there is no lockout globally or on the client directly
     */
    public boolean isRequestAllowed(RequestIdentifier request) {
        return !bruteForceConfig.isBruteForceProtectionEnabled() ||                             // true if bfp disabled
                isTrustedHost(request) ||                                                       // or we already trust the client
                !(bruteForceConfig.isClientLockoutEnabled() && isClientLockedOut(request)) &&   // False if client enabled, and request should reject
                !(bruteForceConfig.isGlobalLockoutEnabled() && isGlobalLockout());              // False if global enabled and request should reject
    }

    public void registerFailure(RequestIdentifier request) {
        // Lets get or create the auth record
        String identifier = request.getIdentifier();

        AuthRecord authRecordCreated = new AuthRecord();
        AuthRecord authRecord = authenticationRecords.putIfAbsent(identifier, authRecordCreated);
        if (authRecord == null) {
            authRecord = authRecordCreated;
        }
        authRecord.newAttempt();

        log.debug("{} has registered as an authentication failure. This is the {} time. They will be locked at {}",
                identifier, authRecord.getFailureCount(), bruteForceConfig.getMaxClientFailures());

        // We should check if trust has been lost from this request
        if (authRecord.getFailureCount() > bruteForceConfig.getMaxClientFailures() && trustedHosts.remove(identifier)) {
            log.debug("{} has been removed from the trusted clients list due to numerous failures", request);
        }
    }

    /**
     * Checks if a client has attempted an incorrect password too many times in the current period
     *
     * @return true if client is not allowed to make requests
     */
    private boolean isClientLockedOut(RequestIdentifier request) {
        AuthRecord authRecord = authenticationRecords.get(request.getIdentifier());

        if (authRecord != null && authRecord.getFailureCount() >= bruteForceConfig.getMaxClientFailures()) {
            log.debug("{} has been locked from making a request due to too many failures recently", request.getIdentifier());
            return true;
        }
        return false;
    }

    /**
     * Checks if the failures in the current period exceed the limit
     */
    private boolean isGlobalLockout() {
        // We now have to clear any records that are old
        performRecordClean();
        int count = 0;
        for (AuthRecord authRecord : authenticationRecords.values()) {
            count += authRecord.getFailureCount();
        }

        if (count >= bruteForceConfig.getMaxGlobalFailures()) {
            if (!globalLockout) {
                // Then we just started a global lockout
                log.debug("Global lockout has started for all non-trusted hosts due to suspected password brute forcing");
            }
            globalLockout = true;
            return true;
        }
        globalLockout = false;
        return false;
    }

    private boolean isTrustedHost(RequestIdentifier request) {
        if (!request.isAnonymous() && trustedHosts.contains(request.getIdentifier())) {
            log.debug("{} is trusted, therefore can continue with request", request.getIdentifier());
            return true;
        }
        return false;
    }

    /**
     * This aims to remove all clients that have not failed a request in the last period. Only must be performed before a
     * global check.
     */
    private void performRecordClean() {
        Calendar instance = Calendar.getInstance();
        instance.setTime(timeProvider.getCurrentTime());
        instance.add(Calendar.MINUTE, -bruteForceConfig.getTimeoutPeriod());
        Date expire = instance.getTime();

        for (Iterator<AuthRecord> iterator = authenticationRecords.values().iterator(); iterator.hasNext(); ) {
            AuthRecord authRecord = iterator.next();
            if (authRecord.getLastFailure().before(expire)) {
                iterator.remove();
            }
        }
    }
}
